/*
 * ForwardFork
 * forwardFork.js
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is ForwardFork code.
 *
 * The Initial Developer of the Original Code is Matthew Waymost.
 * Portions created by the Inital Developer are Copyright (C) 2005
 * Matthew Waymost. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License Version 2 or later (the  "GPL"), in
 * which case the provisions of the GPL are applicable instead of those
 * above. If you wish to allow use of your version of this file only
 * under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice and
 * other provisions required by the GPL. If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the GPL.
 */

var ForwardFork = {

  // session history
  sHistory:		null,

  // tab information
  tabbrowser:		null,
  currentTab:		null,
  curBrowser:		null,

  // preferences tree strings
  prefFdfkRoot:		'forwardfork.',
  prefFdfkOptionTree:	'preferences.',
  prefFdfkHistoryTree:	'history.',

  // services
  prompts:		null,
  prefs:		null,
  mediator:		null,
  observer:		null,

  // settings
  totalForks:		null,
  minPagesPerFork:	null,

  // performs commands on window load
  // - despite not using it when calling the function, you need a parameter for a JS event,
  // - otherwise call to nsIDOMEventListener will be thrown null exception and the JS
  // - breaks
  onLoad: function(event) {

    // register the ForwardFork observer
    ForwardFork.observer = new fdfkObserver();
    ForwardFork.observer.register();

    // context menu listener
    var contextMenu = document.getElementById("contentAreaContextMenu");
    contextMenu.addEventListener("popupshowing", this.onPopupshowing, false);

    // load services
    this.prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Components.interfaces.nsIPromptService);
    this.prefs = Components.classes["@mozilla.org/preferences-service;1"].
                 getService(Components.interfaces.nsIPrefService);
    this.prefs = this.prefs.getBranch(ForwardFork.prefFdfkRoot + ForwardFork.prefFdfkOptionTree);
    this.mediator = Components.classes['@mozilla.org/appshell/window-mediator;1'].
                    getService(Components.interfaces.nsIWindowMediator);

    // load preferences
    ForwardFork.loadPrefs();

    // get tabbrowser container and current browser container with session history
    ForwardFork.tabbrowser = window.getBrowser();
    ForwardFork.currentTab = ForwardFork.tabbrowser.mTabContainer.selectedIndex;
    ForwardFork.curBrowser = ForwardFork.tabbrowser.getBrowserAtIndex(ForwardFork.currentTab);

    // create the fdfkForks array and fdfkGoneBack flag for the current tab
    if (!ForwardFork.curBrowser.fdfkForks) ForwardFork.curBrowser.fdfkForks = new Array();
    if (!ForwardFork.curBrowser.fdfkGoneBack) ForwardFork.curBrowser.fdfkGoneBack = false;

    // lets ForwardFork know when the selected tab changes
    ForwardFork.tabbrowser.addEventListener("select", ForwardFork.onTabChange, false);

    // pull the seesion history, if it exists (it damn well better), then listen to it
    // - BUG: if new window opens to about:blank no history is yet created, listener won't commit
    // - think it has to do with not creating global history upon startup, refer to line 550 of
    //   browser.js
    // - how to allow for this behavior?
    ForwardFork.sHistory = ForwardFork.curBrowser.webNavigation.sessionHistory;
    if (ForwardFork.sHistory) ForwardFork.sHistory.addSHistoryListener(ForwardFork);
  },

  onUnload: function(event) {

    // ForwardFork.savePrefs();
    ForwardFork.observer.unregister();
  },

  loadPrefs: function() {

    // if preference is for some st00pid reason broken, then revert to a hardcoded default
    // - better way? revert to default pref?

    // forwardfork.preferences.maximumforks
    if (ForwardFork.prefs.getPrefType("maximumforks") == ForwardFork.prefs.PREF_INT) {
      ForwardFork.totalForks = ForwardFork.prefs.getIntPref("maximumforks");
    } else {
      ForwardFork.prefs.setIntPref("maximumforks", 6);
      ForwardFork.totalForks = 6;
    }

    // forwardfork.preferences.minimumforksize
    if (ForwardFork.prefs.getPrefType("minimumforksize") == ForwardFork.prefs.PREF_INT) {
      ForwardFork.minPagesPerFork = ForwardFork.prefs.getIntPref("minimumforksize");
    } else {
      ForwardFork.prefs.setIntPref("minimumforksize", 2);
      ForwardFork.minPagesPerFork = 2;
    }
  },

  savePrefs: function() {

    ForwardFork.prefs.setIntPref("maximumforks", ForwardFork.totalForks);
    ForwardFork.prefs.setIntPref("minimumforksize", ForwardFork.minPagesPerFork);
  },

  onTabChange: function(event) {

    // remove listener from previous tab...
    if (ForwardFork.sHistory) ForwardFork.sHistory.removeSHistoryListener(ForwardFork);
    ForwardFork.sHistory = null;

    // ...so we can attach it to the current tab
    ForwardFork.currentTab = ForwardFork.tabbrowser.mTabContainer.selectedIndex;
    ForwardFork.curBrowser = ForwardFork.tabbrowser.getBrowserAtIndex(ForwardFork.currentTab);
    ForwardFork.sHistory = ForwardFork.curBrowser.webNavigation.sessionHistory;
    ForwardFork.sHistory.addSHistoryListener(ForwardFork);

    // and create the fdfkForks array and fdfkGoneBack flag if they don't yet exist
    if (!ForwardFork.curBrowser.fdfkForks) ForwardFork.curBrowser.fdfkForks = new Array();
    if (!ForwardFork.curBrowser.fdfkGoneBack) ForwardFork.curBrowser.fdfkGoneBack = false;
  },

  // Function that controls when fdfk context menu and its submenus are shown
  // - only show ForwardFork menu when history items show up
  // - TO UPDATE: menus hardcoded in XUL overlay; should implement dynamic system so user
  //   may specify as many fork as s/he desires
  onPopupshowing: function() {

    gContextMenu.showItem("fdfk-context", !( gContextMenu.isContentSelected || gContextMenu.onLink || gContextMenu.onImage || gContextMenu.onTextInput ) );
    gContextMenu.showItem("fdfk-context-sep", !( gContextMenu.isContentSelected || gContextMenu.onLink || gContextMenu.onImage || gContextMenu.onTextInput ) );

    gContextMenu.showItem("fdfk-context-no-fork-item" ,!(ForwardFork.showForkMenu(1)));
    gContextMenu.showItem("fdfk-context-fork-1-menu", ForwardFork.showForkMenu(1));
    gContextMenu.showItem("fdfk-context-fork-2-menu", ForwardFork.showForkMenu(2));
    gContextMenu.showItem("fdfk-context-fork-3-menu", ForwardFork.showForkMenu(3));
    gContextMenu.showItem("fdfk-context-fork-4-menu", ForwardFork.showForkMenu(4));
    gContextMenu.showItem("fdfk-context-fork-5-menu", ForwardFork.showForkMenu(5));
    gContextMenu.showItem("fdfk-context-fork-6-menu", ForwardFork.showForkMenu(6));
  },

  // Determines whether a menu should be shown by two criteria:
  // - it is <= the total number of forks
  // - the correspondent fork exists
  showForkMenu: function(num) {

    var inBound = (num <= ForwardFork.totalForks);
    var exists;

    if (ForwardFork.curBrowser.fdfkForks) {
      if (ForwardFork.curBrowser.fdfkForks[num])
        exists = (ForwardFork.curBrowser.fdfkForks[num][0]) ? true : false;
    } else return false;

    return (inBound && exists);
  },

  // Opens or refocuses the options dialog as appropriate
  openOptionsDialog: function() {

    var optionsWindow = ForwardFork.mediator.getMostRecentWindow("forwardfork:options");

    // if its open, focus it; otherwise, open it
    if (optionsWindow && !optionsWindow.closed) optionsWindow.document.commandDispatcher.focusedWindow.focus();
    else optionsWindow = window.openDialog("chrome://forwardfork/content/forwardForkPrefs.xul","ForwardFork Options","chrome,centerscreen,resizable");

    return optionsWindow;
  },

  // buildFork function takes the forward history of the passed through tab and builds
  // a fork from it
  // - derived from browser.js history code
  // - returns true if successful, false otherwise
  buildFork: function(tab, fork, index, end) {

    // dump forward history into awesome forks array!!
    var count = 0; // why doesn't j-index+1 work???
    for (var j=index+1; j<=end; ++j) {
      entry = ForwardFork.sHistory.getEntryAtIndex(j, false);
      if (entry) ForwardFork.curBrowser.fdfkForks[fork][count++] = entry;
    }

    return true;
  },

  // Function that generates the appropriate fork menu on the fly for XUL.
  // - uses standard functions in browser.js
  ForkMenu: function(event, id) {

    var parent = event.target;
    var curTab = ForwardFork.currentTab;

    // naively delete old menu
    // - function in browser.js
    deleteHistoryItems(parent);

    var length = ForwardFork.curBrowser.fdfkForks[id].length;

    for (var i=0; i<length; ++i) {
      createMenuItem(parent, i, ForwardFork.curBrowser.fdfkForks[id][i].title);
    }
  },

  // Function that follows the link selected in fork menu by fork id and
  // entry index.
  // - derived from gotoHistoryIndex(aEvent) in browser.js
  FollowLinkByIndex: function(event, id) {

    var index = event.target.getAttribute("index");
    if (!index) return false;

    // Firefox function that checks for modified clicks and returns where to open
    // the URI
    var where = whereToOpenLink(event);

    var entry = ForwardFork.curBrowser.fdfkForks[id][index];

    // since we don't need to access the history and the built-in function supports
    // where=="cuurent", we can dump the conditional and use the following line to
    // open the link
    openUILinkIn(entry.URI.spec, where);

    return true;
  },

  // *** begin nsISHistoryListener implementation functions
  // - refer to http://www.xulplanet.com/references/xpcomref/ifaces/nsISHistoryListener.html
  //   for specification

  // OnHistoryGoBack
  // - Called to notify a listener when the user presses the 'back' button of the browser
  //   OR when the user attempts to go back one page in history thro' other means, from
  //   javascript or using nsIWebNavigation
  // - Similar function perform similar tasks.
  OnHistoryGoBack: function(backURI) {

    // this.prompts.alert(null, "nsISHistoryListener.OnHistoryGoBack", this.currentTab); // debug

    if (!ForwardFork.curBrowser.fdfkGoneBack) ForwardFork.curBrowser.fdfkGoneBack = true;

    return true;
  },

  OnHistoryGoForward: function(forwardURI) {

    return true;
  },

  OnHistoryGotoIndex: function(index, gotoURI) {

    var curIndex = ForwardFork.sHistory.index;
    if (index < curIndex && !ForwardFork.curBrowser.fdfkGoneBack) {
      ForwardFork.curBrowser.fdfkGoneBack = true;
      // this.prompts.alert(null, "nsISHistoryListener.OnHistoryGotoIndex", this.currentTab + " shift " + index-curIndex); // debug
    }

    return true;
  },

  // OnHistoryNewEntry
  // - Called to notify a listener when a new document is added to session history.
  //   New documents are added to session history by docshell when new pages are loaded
  //   in a frame or content area.
  // * returns void *
  OnHistoryNewEntry: function(newURI) {

    if (ForwardFork.curBrowser.fdfkGoneBack) {

      var curTab = ForwardFork.currentTab;
      var end = ForwardFork.sHistory.count - 1;
      var index = ForwardFork.sHistory.index;

      // check to make sure there's enough history to fork; if not, do nothing
      // - accounts for ForwardFork.minPagesPerFork (index+min) min>=1
      if ((index + ForwardFork.minPagesPerFork) > end) {;}
      else {
        if (!ForwardFork.curBrowser.fdfkForks[0]) ForwardFork.curBrowser.fdfkForks[0] = 1;

        // retrieve the highest empty fork from correct tab
        // - the maximum of highestFork is totalForks, not totalForks+1
        //   since when dropping down, the highest slot can be considered
        //   empty since the fork is lost
        var highestFork = ForwardFork.curBrowser.fdfkForks[0];

        // drop down forks
        var track = highestFork;
        while (track > 1) {
          ForwardFork.curBrowser.fdfkForks[track] = ForwardFork.curBrowser.fdfkForks[track-1];
          --track;
        }
        ForwardFork.curBrowser.fdfkForks[1] = new Array();

        // build new fork from history and update fork information if necessary
        if (ForwardFork.buildFork(curTab, 1, index, end)) {
          ForwardFork.curBrowser.fdfkForks[0] = (highestFork < ForwardFork.totalForks) ? ++highestFork : ForwardFork.totalForks;
        }

        // even if absolute disaster occurs, don't try again
        ForwardFork.curBrowser.fdfkGoneBack = false;
      }
    }
  },

  OnHistoryPurge: function(numEntries) { return true; },

  OnHistoryReload: function(reloadURI, reloadFlags) { return true; },

  // *** end nsISHistoryListener implementation functions

  // QueryInterface function so XPCOM know what the hell the ForwardFork object
  // implements
  QueryInterface: function(aIID) {

    if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
        aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
        aIID.equals(Components.interfaces.nsISupports) )
      return ForwardFork;
    throw Components.results.NS_NOINTERFACE;
  }
};

// ForwardFork observer
// - listens for message from prefs dialog and modifies settings appropriately
// - refer to MozZine KB for details
function fdfkObserver() { this.register(); }

fdfkObserver.prototype = {

  observe: function(subject, topic, data) {

    var changed = true;

    if (topic == "forwardfork:minforksize") ForwardFork.minPagesPerFork = data;
    else if (topic == "forwardfork:maxforks") ForwardFork.totalForks = data;
    else changed = false;

    if (changed) ForwardFork.savePrefs();
  },

  register: function() {
    var observerService = Components.classes["@mozilla.org/observer-service;1"]
                          .getService(Components.interfaces.nsIObserverService);
    observerService.addObserver(this, "forwardfork:maxforks", false);
    observerService.addObserver(this, "forwardfork:minforksize", false);
  },

  unregister: function() {
    var observerService = Components.classes["@mozilla.org/observer-service;1"]
                          .getService(Components.interfaces.nsIObserverService);
    observerService.removeObserver(this, "forwardfork:maxforks");
    observerService.removeObserver(this, "forwardfork:minforksize");
  }
}

// Accessor functions
function fdfkOnLoad() { ForwardFork.onLoad(); }
function fdfkOnUnload() { ForwardFork.onUnload(); }

// Load event listeners
window.addEventListener('load', fdfkOnLoad, false);
window.addEventListener('unload', fdfkOnUnload, false);